/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: satyr.c,v 1.6 2000/08/10 11:22:00 bazsi Exp $
 *
 ***************************************************************************/

#include <zorp/satyr.h>
#include <zorp/authconn.h>
#include <zorp/labelset.h>
#include <zorp/log.h>
#include <zorp/lineio.h>

#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>

#include <string.h>

#include <openssl/ssl.h>
#include <openssl/err.h>
#include <glib.h>

#define SATYR_PORT (0xaffe)

char *z_satyr_privkey_file = ZORP_SYSCONFDIR "/zorp.key";
char *z_satyr_cert_file = ZORP_SYSCONFDIR "/zorp.crt";
char *z_satyr_ca_file = ZORP_SYSCONFDIR "/ca.crt";

/*                                                                                          
extern char *z_satyr_privkey_file;
extern char *z_satyr_cert_file;
extern char *z_satyr_ca_file;
*/

typedef struct _ZRealSatyrAuthenticator
{
  guint ref_count;
  ZAuthConnection *auth_conn;
  SSL_CTX *ssl_ctx;
}
ZRealSatyrAuthenticator;

/*
 *  Creates a new ZSatyrAuthenticator, and sets up a new SSL contex for it.
 *  On error it returns NULL. Else the pointer to the newly allocated
 *  ZSatyrAuthenticator.
 */
ZSatyrAuthenticator *
z_satyr_authenticator_new (ZAuthConnection * auth_conn)
{
  ZRealSatyrAuthenticator *self;
  SSL_METHOD *ssl_meth;
  SSL_CTX *ctx;

  g_return_val_if_fail (auth_conn, NULL);

  self = g_new0 (ZRealSatyrAuthenticator, 1);

  if (!self)
    return NULL;

  self->auth_conn = auth_conn;

  SSL_load_error_strings ();
  OpenSSL_add_all_algorithms();

  ssl_meth = SSLv3_server_method ();
  self->ssl_ctx = ctx = SSL_CTX_new (ssl_meth);

  if (!ctx)
    {
      z_warning (0, "Error creating SSL context: %s",
		 ERR_error_string (ERR_get_error (), NULL));
      g_free (self);
      return NULL;
    }

  if (SSL_CTX_use_PrivateKey_file
      (ctx, z_satyr_privkey_file, SSL_FILETYPE_PEM) < 0
      || SSL_CTX_use_certificate_file (ctx, z_satyr_cert_file,
				       SSL_FILETYPE_PEM) < 0
      || !SSL_CTX_check_private_key (ctx))
    {
      z_warning (0, "Error loading keys: %s",
		 ERR_error_string (ERR_get_error (), NULL));
      SSL_CTX_free (ctx);
      g_free (self);
      return NULL;
    }

  SSL_CTX_set_verify (ctx, SSL_VERIFY_NONE, NULL);

  self->ref_count = 1;

  return (ZSatyrAuthenticator *) self;
}

/*
 *  destroys the ZSatyrAuthenticator.
 *  It is invoked by z_satyr_authenticator_unref when needed.
 */
static void
z_satyr_authenticator_destroy (ZSatyrAuthenticator * s)
{
  ZRealSatyrAuthenticator *self = (ZRealSatyrAuthenticator *) s;

  g_return_if_fail (s);

  SSL_CTX_free (self->ssl_ctx);
  g_free (self);
}

/*
 *  increases the ref_num in the ZSatyrAuthenticator
 */
void
z_satyr_authenticator_ref (ZSatyrAuthenticator * s)
{
  ZRealSatyrAuthenticator *self = (ZRealSatyrAuthenticator *) s;

  g_return_if_fail (s);

  self->ref_count++;
}

/*
 *  Decreases the ref_num in a ZSatyrAuthenticator, and destroys it, if needed.
 */
void
z_satyr_authenticator_unref (ZSatyrAuthenticator * s)
{
  ZRealSatyrAuthenticator *self = (ZRealSatyrAuthenticator *) s;

  g_return_if_fail (s);

  self->ref_count--;
  if (self->ref_count == 0)
    z_satyr_authenticator_destroy (s);
}

/*
 * Creates an internet connection to the given IP address on the port defined
 * by SATYR_PORT. Returns the connected fd, or -1 on error.
 */
static int
z_satyr_make_inet_connection (guint ip)
{
  struct sockaddr_in sn;
  int sock, ret;

  sn.sin_family = AF_INET;
  sn.sin_addr.s_addr = htonl (ip);
  sn.sin_port = htons (SATYR_PORT);

  sock = socket (PF_INET, SOCK_STREAM, 0);
  if (sock < 0)
    {
      z_warning (3, "Cannot create socket %s", strerror (errno));
      return -1;
    }

  ret = connect (sock, (struct sockaddr *) &sn, sizeof (sn));
  if (ret < 0)
    {
      z_warning (3, "Cannot connect %s", strerror (errno));
      return -1;
    }

  return sock;
}

/*
 *  Sends the first, authenticatee selector command to the satyr client,
 *  (on unix satyrmpxd) Returns 0 on success, and negative on error.
 */
static int
z_satyr_choose_user (int fd, guint type, guint port)
{
  char buffer[16];

  snprintf (buffer, 15, "%s/%05d\n",
	    (type == Z_SATYR_TYPE_UDP ? "UDP" : "TCP"), port);
  write (fd, buffer, 10);
  read (fd, buffer, 4);
  buffer[4] = 0;
  if (!strncmp (buffer, "ACK\n", 4))
    return 0;
  if (!strncmp (buffer, "ERR\n", 4))
    return -1;
  return -2;
}

/*
 *  Creates a new ssl connection on 'fd', useing 'ctx' as the SSL context.
 *  Returns the newly allocated SSL structure, or NULL on error.
 */
static SSL *
z_satyr_make_ssl_connection (SSL_CTX * ctx, int fd)
{
  SSL *ssl;

  g_return_val_if_fail (ctx, NULL);

  ssl = SSL_new (ctx);
  if (!ssl)
    {
      z_warning (0, "Error creating SSL structure: %s",
		 ERR_error_string (ERR_get_error (), NULL));
      return NULL;
    }

  SSL_set_fd (ssl, fd);

  if (SSL_accept (ssl) < 0)
    {
      z_warning (0, "Error accepting SSL connection: %s",
		 ERR_error_string (ERR_get_error (), NULL));
      SSL_free (ssl);
      return NULL;
    }

  return ssl;
}

/*
 *  Theese are the states of the authenticator state-machine
 */
#define Z_SATYR_STATE_TRAP           0
#define Z_SATYR_STATE_INIT           1
#define Z_SATYR_STATE_BASE           2
#define Z_SATYR_STATE_USERNAME       3
#define Z_SATYR_STATE_METHOD         4
#define Z_SATYR_STATE_RESPONSE       5

/*
 *  The followings are command sets, which says which commands may be used
 *  in a state of the protocol.
 */
static Command user_cmds[] = {
  {"NOAUTH", 0},
  {"USER", 1},
  {NULL, 0}
};

static Command meth_cmds[] = {
  {"NOAUTH", 0},
  {"USER", 1},
  {"CHME", 1},
  {NULL, 0}
};

static Command resp_cmds[] = {
  {"NOAUTH", 0},
  {"USER", 1},
  {"CHME", 1},
  {"RESP", 1},
  {NULL, 0}
};

#define Z_SATYR_CMD_NOAUTH   0
#define Z_SATYR_CMD_USER     1
#define Z_SATYR_CMD_CHME     2
#define Z_SATYR_CMD_RESP     3

/*
 *  Calculates which methods may be used. It takes a list of useable methods
 *  in (*methods), which are defined by the policy. It also takes a username
 *  and asks the authentication server for the available methods for the 
 *  specific user. Then takes the intercept of the two lists.
 */
static void
z_satyr_calculate_methods (ZSatyrAuthenticator * s,
			   GList ** methods, char *username)
{
  ZRealSatyrAuthenticator *self = (ZRealSatyrAuthenticator *) s;
  char **usrmethods;
  int methnum, i;
  GList *p;
  GList *old_method_list = *methods;
  GList *new_method_list = NULL;

  usrmethods = z_authconn_utyp (self->auth_conn, username, &methnum);

  for (i = 0; i < methnum; i++)
    {
      p = old_method_list;
      while (p)
	{
	  if (!strcmp (p->data, usrmethods[i]))
	    {
	      new_method_list = g_list_append (new_method_list, p->data);
	      old_method_list = g_list_remove (old_method_list, p->data);
	      break;
	    }
	  p = p->next;
	}
    }

  g_list_free (old_method_list);

  *methods = new_method_list;

  return;
}

/*
 *  Gets a challange for user 'username' for the authentication method named
 *  in 'method' from the authentication server. Returns it as a newly allocated
 *  string which should be freed later by the caller by a g_free.
 *  If no challange is available it return NULL.
 */
static char *
z_satyr_get_challange (ZSatyrAuthenticator * s,
		       char *method, char *username)
{
  ZRealSatyrAuthenticator *self = (ZRealSatyrAuthenticator *) s;
  char buffer[1024];
  char *ret;

  if(z_authconn_chal (self->auth_conn, method, username, buffer, 1023)){
    ret = g_new0 (char, strlen(buffer)+1);
    strcpy(ret,buffer);
  }else{
    ret = NULL;
  }

  return ret;
}


/*
 *  Authenticates the user using the given parameters to the authentication
 *  server. Returns exactly what the auth_server returns.
 */
static ZLabelSet *
z_satyr_real_authenticate (ZSatyrAuthenticator * s,
			   char *username,
			   char *method,
			   char *chal, char *response)
{
  ZRealSatyrAuthenticator *self = (ZRealSatyrAuthenticator *) s;

  return z_authconn_auth (self->auth_conn, method, username, chal, response);
}


/*
 *  Processes a complete authentication procedure. Connects to the client,
 *  chooses the appropriate user, and talks to satyrd useing the correct
 *  protocol.
 */
ZLabelSet *
z_satyr_authenticator_auth (ZSatyrAuthenticator * s,
			    ZSatyrSubjID * subj,
			    char *object_descr, char **methods)
{
  ZRealSatyrAuthenticator *self = (ZRealSatyrAuthenticator *) s;
  int fd, ret;
  SSL *ssl;
  GList *usable_methods = NULL;
  ZLabelSet *return_value = NULL;
  char **p;
  
// allapotgep valtozoi
  char *username = NULL, *method = NULL, *chal = NULL;
  Command *cmd_set;		// melyik parancsok johetnek.
  int state;
  int retry;
  
  GList *line;
  int which, argnum;


  g_return_val_if_fail (s, NULL);
  g_return_val_if_fail (subj, NULL);
  g_return_val_if_fail (object_descr, NULL);
  g_return_val_if_fail (methods, NULL);


  fd = z_satyr_make_inet_connection (subj->ip);
  if (fd < 0)
    return NULL;

  ret = z_satyr_choose_user (fd, subj->type, subj->port);
  if (ret < 0)
    {
      close (fd);
      return NULL;
    }

  ssl = z_satyr_make_ssl_connection (self->ssl_ctx, fd);
  if (!ssl)
    {
      close (fd);
      return NULL;
    }

  state = Z_SATYR_STATE_INIT;
  cmd_set = user_cmds;
  retry=2;
  while (state)
    {
      switch (state)
	{
	case Z_SATYR_STATE_INIT:
	  write_token_line (ssl, "ss", "AUTH", object_descr);
	  state = Z_SATYR_STATE_BASE;
	  break;
	case Z_SATYR_STATE_BASE:
	  line = read_command (ssl, cmd_set, &which, &argnum);
	  if (!line)
	    {
	      z_message (5, "Satyr peer failed using the protocol");
	      state = Z_SATYR_STATE_TRAP;
	      break;
	    }
	  switch (which)
	    {
	    case Z_SATYR_CMD_NOAUTH:
	      state = Z_SATYR_STATE_TRAP;
	      break;
	    case Z_SATYR_CMD_USER:
	      state = Z_SATYR_STATE_USERNAME;
	      break;
	    case Z_SATYR_CMD_CHME:
	      state = Z_SATYR_STATE_METHOD;
	      break;
	    case Z_SATYR_CMD_RESP:
	      state = Z_SATYR_STATE_RESPONSE;
	      break;
	    default:
	      z_warning (5, "z_satyr: there is some error...");
	      state = Z_SATYR_STATE_TRAP;
	    }
	  break;
	case Z_SATYR_STATE_USERNAME:
	  if (username)
	    g_free (username);
	  username = g_new0 (char, strlen (line->next->data) + 1);
	  strcpy (username, line->next->data);
	  free_g_list_of_strings (line);

    p=methods;
    while (*p)
      {
        usable_methods = g_list_append (usable_methods, *p);
        p++;
      }

	  z_satyr_calculate_methods (s, &usable_methods, username);
	  if (!usable_methods)
	    write_token_line (ssl, "s", "METH");
	  else
	    write_token_line (ssl, "sg", "METH", usable_methods);

	  cmd_set = meth_cmds;
	  state = Z_SATYR_STATE_BASE;
	  break;
	case Z_SATYR_STATE_METHOD:
	  if (method)
	    g_free (method);
	  method = g_new0 (char, strlen (line->next->data) + 1);
	  strcpy (method, line->next->data);
	  free_g_list_of_strings (line);

	  if (chal)
	    g_free (chal);
	  chal = z_satyr_get_challange (s, method, username);
    
    if(chal){
	    write_token_line (ssl, "ss", "CHAL", chal);
  	  cmd_set = resp_cmds;
    }else
      write_token_line(ssl,"s","CHAL");

	  state = Z_SATYR_STATE_BASE;
	  break;
	case Z_SATYR_STATE_RESPONSE:
	  return_value =
	    z_satyr_real_authenticate (s, username, method, chal,
				       line->next->data);
	  free_g_list_of_strings (line);

	  if (return_value)
	    write_token_line (ssl, "s", "ACCEPT");
	  else{
	  	if(retry--){
		  write_token_line(ssl,"s","RETRY");
		  state = Z_SATYR_STATE_BASE;
		  break;
		}else
	    	write_token_line (ssl, "s", "REJECT");
	  }

	  state = Z_SATYR_STATE_TRAP;
	  break;
	case Z_SATYR_STATE_TRAP:
	  break;
	default:
	  z_warning (5, "z_satyr_authenticator: unknown state");
	}
    }

  if (username)
    g_free (username);
  if (method)
    g_free (method);
  if (chal)
    g_free (chal);

  SSL_shutdown (ssl);
  close (fd);
  SSL_free (ssl);
  return return_value;
}
