############################################################################
##
## Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: Chainer.py,v 1.18 2000/08/08 14:39:00 bazsi Exp $
##
############################################################################

from Zorp import *
import Connector, Stream, SockAddr
import socket

class Chainer:
	"""
		Ez az osztaly hvdik meg, ha valaki hozzcsatlakozik a proxy-hoz,
		itt pl fel a kapcsolat a szerver oldalal.
	"""
	def connectServer(self, session, host):
		raise NotImplementedError
		
	def establishConnection(self, session, local, remote):
		"""
			Ez a fggvny hajtja vgre a csatlakozst, ahhoz a serverhez, amit a
			connectServer fgv. meghatrozott

			Paramterek:
				remote			A server adatai, amihez csatlakozni kell.

			Lers:
				Ez a fggvny ltrehozza a csatlakozst a proxy-tol a
				megadott szerverig.

			Visszatrsi rtk:
				int				A szerver fel men csatorna fd-je.
		"""
		try:
			session.setServer(remote)
			conn = Connector.Connect(local, remote)
			session.server_fd = conn.blockingConnect()
			if session.server_fd != -1:
				session.server = Stream.Stream(session.server_fd, session.session_id + ": server")
				return session.server_fd
		except DACException:
			pass
		except MACException:
			pass
		return -1

		

class DirectedChainer(Chainer):
	"""
		Ez a csatlakozs elre meghatrozott szerverhez csatlakozik,
		brmely krs esetn

		local						A csatlkozs loklis adatai
		remote					A szerver adatai
	"""
	local = None
	remote = None
	
	def __init__(self, local, remote):
		"""
			Ltrehoz egy Chainer objektumot.

			Paramterek:
				local				A csatlakozs loklis adatai
				remote			A szerver adatai, amihez csatlakozni kell

			Lers:
				Belltja a loklis vltozkat a megadott paramterek
				alapjn

			Visszatrsi rtk:
				-
		"""
		self.local = local
		self.remote = remote
	
	def connectServer(self, session, host, port):
		"""
			Ez a fuggvny hvodik meg kvlrl csatlakozs esetn, s dnti el,
			hogy hova is kell a proxy-nak csatlakoznia

			Paramterek:
				session			Az aktulis kapcsols adatai
				host				Az tadott gepcm, ??Nincs hasznlatban.??
				port				Az tadott portcm. ??Nincs hasznlatban.??

			Lers:
				Elinditja az establishConnection-t a beltott rtkeknek
				megfelelen.

			Visszatrsi rtk:
				int				Az establishConnect lltal visszaadott fd.
		"""
		return self.establishConnection(session, self.local, self.remote)

class TransparentChainer(Chainer):
	"""
		Ez a csatlakozs kapcsolatonknt ms, a kapcsolat lltal meghatrozott
		helyhez csatlakozik.

		Vltozk:
			forge_addr			trja-e a kliens adatait egy megadottra
									csatlakozskor
			local					Ha igen, akkor mire.
	"""
	forge_addr = 0
	local = None
	
	def __init__(self, local, forge_addr):
		"""
			Ltrehozza az osztly egy pldnyt.

			Paramterek:
				local				Mire rja t a klien adatait NINCS HASZNALVA
				forge_addr		trja-e a kliens adatait

			Lers:
				Ltrehozza az osztly egy pldnyt, s a paramternek megfelelen
				belltja, hogy trdjanak-e a klien adatai

			Visszatrsi rtk:
				-
		"""
		self.forge_addr = forge_addr
	
	def connectServer(self, session, host, port):
		"""
			Ez a fuggvny hvodik meg kvlrl csatlakozs esetn, s dnti el,
			hogy hova is kell a proxy-nak csatlakoznia

			Paramterek:
				session			Az aktulis kapcsols adatai
				host				Az tadott gpcm. ??Nincs hasznlatban??
				port				Az tadott portcm. ??Nincs hasznlatban??

			Lers:
				Csatlakozik a session alltal meghatarozott helyre, opcionlisan
				elrejtve a kliens adatait.

			Visszatrsi rtk:
				int				Az establishConnect lltal visszaadott fd.
		"""

		if self.forge_addr:
			local = session.client_addr
		else:
			local = self.local
		remote = getTransparentDestination(session.client_fd)
		return self.establishConnection(session, self.local, remote)

class InbandChainer(Chainer):
	local = None
	
	def __init__(self, local):
		self.local = local
	
	def connectServer(self, session, host, port):
		#message(0, "InbandChainer.connectServer()")
		try:
			ip = socket.gethostbyname(host)
			return self.establishConnection(session, self.local, SockAddr.SockAddrInet(ip, port))
		except socket.error:
			return -1

class FailoverChainer(Chainer):
	local = None
	hosts = []
	current_host = 0
	
	def __init__(self, local, hosts):
		self.local = local
		self.hosts = hosts
	
	def nextHost(self):
		self.current_host = self.current_host + 1
		if self.current_host == len(self.hosts):
			self.current_host = 0

	def getHostAddress(self):
		return self.hosts[self.current_host]
	
	def connectServer(self, session, host, port):
		first_attempt = self.current_host
		remote = self.getHostAddress()
		res = self.establishConnection(session, self.local, remote)
		self.nextHost()
		while res == -1 and self.current_host != first_attempt:
			remote = self.getHostAddress()
			res = self.establishConnection(session, self.local, remote)
			self.nextHost()
		return res
