############################################################################
##
## Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: Zone.py,v 1.11 2000/08/01 10:10:37 bazsi Exp $
##
############################################################################

from Zorp import *
from SockAddr import *

PERMIT_INBOUND	= 1
PERMIT_OUTBOUND = 2

class Zone:
	"""
		Egy csomag (kapcsolat) znjt leir osztly

		Vltozok:
			name			Az egyes pldnyok neve
	"""
	name = ""
	permitted_services = {}
	
	def __init__(self, name, outbound_services=None, inbound_services=None):
		"""
			Ltrehoz egy zna pldnyt

			Paramterek:
				name			A pldny neve

			Lers:
				Bellitja a pldny name vltozjt a megadott paramternek
				megfelelen.

			Visszatrsi rtk:
				-
		"""
		self.name = name
		if outbound_services != None:
			for i in outbound_services:
				self.permitted_services[i] = PERMIT_OUTBOUND
				#debug(0, "zone(%s): outbound: %s" % (self.name, i));
	
		if inbound_services != None:		
			for i in inbound_services:
				try:
					self.permitted_services[i] = self.permitted_services[i] | PERMIT_INBOUND
					#debug(0, "zone(%s): inbound: %s" % (self.name, i));
				except KeyError:
					self.permitted_services[i] = PERMIT_INBOUND
					#debug(0, "zone(%s): inbound: %s" % (self.name, i));
		
		
	def isMember(self, addr, interface = None):
		"""
			Megnzi egy cmrl, hogy ebbe a znba tartozik-e.

			Paramterek
				addr		A csomag cmt tartalmazza
				interface	A fogad interface neve.

			Lers:
				A fuggvny nincs implementlva

			Visszatrsi rtk:
				-
		"""
		raise NotImplementedError
		
	def isInboundServicePermitted(self, session):
		"""
			Eldnti, hogy egy bejv kapcsolat engedlyezhet-e.

			Paramterek:
				session		A kapcsolat lersa

			Lers:
				A fggvny engedlyezi a bejv kapcsolatot.

			Visszatrsi rtk:
				Z_ACCEPT
		"""
		try:
			default = self.permitted_services["*"]
		except KeyError:
			default = 0
		
		try:
			t = self.permitted_services[session.service.name]
			if t & PERMIT_INBOUND:
				return Z_ACCEPT
		except KeyError:
			if default & PERMIT_INBOUND:
				return Z_ACCEPT
			
		return Z_REJECT
		
	def isOutboundServicePermitted(self, session):
		"""
			Eldnti, hogy egy kimen kapcsolat engedlyezhet-e.

			Paramterek:
				session		A kapcsolat lersa

			Lers:
				A fggvny engedlyezi a kimen kapcsolatot.

			Visszatrsi rtk:
				Z_ACCEPT
		"""
		try:
			default = self.permitted_services["*"]
		except KeyError:
			default = 0

		try:
			t = self.permitted_services[session.service]
			if t & PERMIT_OUTBOUND:
				return Z_ACCEPT
		except KeyError:
			if default & PERMIT_OUTBOUND:
				return Z_ACCEPT
			
		return Z_REJECT

class InetZone(Zone):
	"""
	Egy internetrol (IP kapcsolatkent) erkezo csomag znjt ler
	osztly (A Zone osztlybl szrmaztatva)

	Sajat vltozk:
		baseip				A zna IP osztlya (32 bites egsz) (NETWORK)
		mask				A zna maskja. (32 bites egsz)
		interface			A zna fel nz interface
	"""
	baseip = 0
	mask = 0
	interface = None
	
	def __init__(self, name, ip, mask, interface, outbound_services=None, inbound_services=None):
		"""
			Ltrehoz egy InetZone pldnyt.

			Paramterek:
				name			A pldny neve
				ip			A Zna IP osztlya (NETWORK)
							(ponttal elvalasztott 4 szm sztringben)
							(elg ebrmelyik IP szm a znbl)
				mask			A zna maskja. (ponttal elvlasztott 4 szm
							sztringben)
				interface		A zna fel nz interface
				permitted_services	A znbl kifel in

			Lers:
				Ltrehoz egy InetZone pldnyt, belltva a tartomny
				jellemzit

			Visszetrsi rtk:
				-
		"""
		Zone.__init__(self, name, outbound_services, inbound_services)
		self.mask = inet_aton(mask)
		self.baseip = inet_aton(ip) & self.mask
		self.interface = interface

	def isMember(self, addr, interface = None):
		"""
			Visszaadja, hogy egy csomag beletartozik-e ebbe a znba

			Paramterek:
				addr		A csomag jellemzi
				interface	A csomag interface-e (alaprtelmezett: brmelyik)

			Lers:
				Megnzi, hogy a megadott csomag ip cime s a zna netmask-jnak
				logikai s mvelete utn a zna network szmt kapja-e, s
				ha a znban van definilva az interface, akkor az megegyezik-e
				a csomag interface-evel.

			Visszatrsi rtk:
				boolean
		"""
		#print "%s.isMember(), addr=%s, base=%x, addrbase=%x" % (self.name, addr.format(), self.baseip, addr.ip & self.mask)
		#print addr.ip
		return (((addr.ip & self.mask) == self.baseip)\
			and (self.interface == None or interface == self.interface))
